import json
import responses
import hmac
import hashlib
import six
from uuid import UUID

from django.contrib.auth.models import User
from django.test import TestCase

from wapipelines.models import (
    Pipeline,
    ImpossiblePipelineException,
    Step)


ARTEFACT_1 = 'artefact-1'
ARTEFACT_2 = 'artefact-2'
ARTEFACT_3 = 'artefact-3'
ARTEFACT_4 = 'artefact-4'


class PipelineTest(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(
            'username', 'test@example.org', 'password')
        self.pipeline = Pipeline.objects.create(
            user=self.user,
            name='test pipeline')

    def mk_simple_steps(self, pipeline,
                        url='http://example.org/',
                        timeout=Step.ACCEPTABLE_TIMEOUT):
        step1 = pipeline.step_set.create(
            url=url,
            name="step1",
            timeout=timeout,
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT_1])
        step2_1 = pipeline.step_set.create(
            url=url,
            name="step2_1",
            timeout=timeout,
            inputs=step1.outputs,
            outputs=[ARTEFACT_2])
        step2_2 = pipeline.step_set.create(
            url=url,
            name="step2_2",
            timeout=timeout,
            inputs=step1.outputs,
            outputs=[ARTEFACT_3])
        step3 = pipeline.step_set.create(
            url=url,
            name="step3",
            timeout=timeout,
            inputs=step2_1.outputs + step2_2.outputs,
            outputs=[ARTEFACT_4])
        step4 = pipeline.step_set.create(
            url=url,
            name="step4",
            timeout=timeout,
            inputs=step3.outputs,
            outputs=Step.EMPTY_OUTPUTS)
        return (step1, step2_1, step2_2, step3, step4)

    def mk_impossible_steps(self, pipeline, url='http://example.org/'):
        step1 = pipeline.step_set.create(
            url=url,
            inputs=[ARTEFACT_1],
            outputs=[ARTEFACT_2])
        step2 = pipeline.step_set.create(
            url=url,
            inputs=[ARTEFACT_2],
            outputs=[ARTEFACT_1])
        return (step1, step2)

    def test_planning(self):
        steps = self.mk_simple_steps(self.pipeline)
        step1, step2_1, step2_2, step3, step4 = steps
        graph = Pipeline.plan(self.pipeline)
        self.assertEqual(
            set(graph.nodes()),
            set(steps))

    def test_detect_impossible_plan(self):
        step1, step2 = self.mk_impossible_steps(self.pipeline)
        self.assertRaises(
            ImpossiblePipelineException,
            Pipeline.plan, self.pipeline)

    def test_find_cycles(self):
        step1, step2 = self.mk_impossible_steps(self.pipeline)
        [cycle] = self.pipeline.get_cycles()
        self.assertEqual(
            set(cycle),
            set([step1, step2]))

    def test_save_planning(self):
        self.mk_simple_steps(self.pipeline)
        graph = Pipeline.plan(self.pipeline)
        self.pipeline.set_plan(graph)
        self.pipeline.save()

        reloaded = Pipeline.objects.get(pk=self.pipeline.pk)
        self.assertEqual(
            reloaded.dot,
            self.pipeline.to_pydot().to_string())

    @responses.activate
    def test_run(self):
        payload = {
            ARTEFACT_1: "value for artefact-1",
            ARTEFACT_2: "value for artefact-2",
            ARTEFACT_3: "value for artefact-3",
            ARTEFACT_4: "value for artefact-4",
        }
        responses.add(
            responses.POST, 'http://example.org/',
            content_type='application/json',
            json=payload)

        steps = self.mk_simple_steps(self.pipeline)
        plan = Pipeline.plan(self.pipeline)
        self.pipeline.set_plan(plan)

        self.pipeline.run({})

        results = self.pipeline.result_set.all()
        # We should run each step only once for each pipeline run
        self.assertEqual(results.count(), len(steps))
        for index, result in enumerate(results):
            self.assertEqual(result.step.pipeline, self.pipeline)
            if index == 0:
                self.assertEqual(result.inputs, {})
            else:
                self.assertEqual(
                    result.inputs.keys(), payload.keys())

            self.assertEqual(result.outputs, payload)
            self.assertEqual(result.response_code, 200)
            self.assertTrue(result.duration.total_seconds())

    @responses.activate
    def test_run_in_parallel(self):

        responses.add(
            responses.POST, 'http://example.org/step1/',
            content_type='application/json',
            json={
                ARTEFACT_1: 'value-for-step-1'
            })
        responses.add(
            responses.POST, 'http://example.org/step2/',
            content_type='application/json',
            json={
                ARTEFACT_2: 'value-for-step-2'
            })
        responses.add(
            responses.POST, 'http://example.org/step3/',
            content_type='application/json',
            json={
                ARTEFACT_3: 'value-for-step-3'
            })
        responses.add(
            responses.POST, 'http://example.org/step4/',
            content_type='application/json',
            json={
                ARTEFACT_4: 'value-for-step-4'
            })

        self.pipeline.step_set.create(
            url='http://example.org/step1/',
            name="step1",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT_1])
        self.pipeline.step_set.create(
            url='http://example.org/step2/',
            name="step2",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=[ARTEFACT_1],
            outputs=[ARTEFACT_2])
        self.pipeline.step_set.create(
            url='http://example.org/step3/',
            name="step3",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=[ARTEFACT_1],
            outputs=[ARTEFACT_3])
        self.pipeline.step_set.create(
            url='http://example.org/step4/',
            name="step4",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=[ARTEFACT_2, ARTEFACT_3],
            outputs=[ARTEFACT_4])

        self.pipeline.run({})

        [run] = self.pipeline.pipelinerun_set.all()
        results = self.pipeline.result_set.all()
        self.assertEqual(
            results.count(),
            4)
        outputs = self.pipeline.get_outputs(run)
        self.assertEqual(outputs, {
            ARTEFACT_1: ['value-for-step-1'],
            ARTEFACT_2: ['value-for-step-2'],
            ARTEFACT_3: ['value-for-step-3'],
            ARTEFACT_4: ['value-for-step-4'],
        })

    @responses.activate
    def test_log_error(self):

        def explode(request):
            return (503, {}, '{}')

        responses.add_callback(
            responses.POST, 'http://example.org/step1/',
            callback=explode,
            content_type='application/json')

        self.pipeline.step_set.create(
            url='http://example.org/step1/',
            name="step1",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT_1])

        self.pipeline.run({})

        [run] = self.pipeline.pipelinerun_set.all()
        [result] = self.pipeline.result_set.all()
        self.assertTrue('503 Server Error' in result.error)
        self.assertEqual(result.response_code, 503)

    @responses.activate
    def test_signing(self):

        def explode(request):
            signing = hmac.new(
                key=six.b(str(self.pipeline.secret)),
                msg=six.b(str(request.body)),
                digestmod=hashlib.sha512)

            self.assertEqual(
                request.headers['X-WA-Sign'],
                signing.hexdigest())

            return (202, {}, '{}')

        responses.add_callback(
            responses.POST, 'http://example.org/step1/',
            callback=explode,
            content_type='application/json')

        self.pipeline.step_set.create(
            url='http://example.org/step1/',
            name="step1",
            timeout=Step.ACCEPTABLE_TIMEOUT,
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT_1])

        self.pipeline.run({})
        self.assertEqual(len(responses.calls), 1)

    def test_entry_points(self):
        step1, step2_1, step_2_2, step3, step4 = self.mk_simple_steps(
            self.pipeline)

        [entrypoint] = self.pipeline.entry_points()
        self.assertEqual(entrypoint, step1)

    def test_exit_points(self):
        step1, step2_1, step_2_2, step3, step4 = self.mk_simple_steps(
            self.pipeline)

        [exitpoint] = self.pipeline.exit_points()
        self.assertEqual(exitpoint, step4)

    def test_connect_pipelines(self):
        pipeline1 = Pipeline.objects.create(
            user=self.user,
            name='pipeline1')
        step1_1 = pipeline1.step_set.create(
            url='http://www.example.org',
            name='step 1 for pipeline 1',
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT_1])
        step1_2 = pipeline1.step_set.create(
            url='http://www.example.org',
            name='step 2 for pipeline 1',
            inputs=[ARTEFACT_1],
            outputs=[ARTEFACT_2])

        pipeline2 = Pipeline.objects.create(
            user=self.user,
            name='pipeline2')
        step2_1 = pipeline2.step_set.create(
            url='http://www.example.org',
            name='step 1 for pipeline 2',
            inputs=[ARTEFACT_2],
            outputs=[ARTEFACT_3])
        step2_2 = pipeline2.step_set.create(
            url='http://www.example.org',
            name='step 2 for pipeline 2',
            inputs=[ARTEFACT_3],
            outputs=[ARTEFACT_4])

        self.assertEqual(pipeline1.entry_points(), [step1_1])
        self.assertEqual(pipeline1.exit_points(), [step1_2])
        self.assertEqual(pipeline2.entry_points(), [step2_1])
        self.assertEqual(pipeline2.exit_points(), [step2_2])

        self.assertEqual(
            pipeline2.can_receive(pipeline1),
            [(step1_2, step2_1)])
        self.assertEqual(
            pipeline1.can_receive(pipeline2),
            [])

    @responses.activate
    def test_submit_outputs(self):

        responses.add(
            responses.POST, 'http://example.org/step/',
            content_type='application/json',
            json={"foo": "bar"})

        def cb(request):
            data = json.loads(request.body)
            self.assertEqual(data["pipeline"]["outputs"], {
                "foo": ["bar"]
            })
            self.assertEqual(
                UUID(data["pipeline"]["uuid"]),
                self.pipeline.uuid)
            return (202, {}, "{}")

        responses.add_callback(
            responses.POST, 'http://example.org/endpoint/',
            content_type='application/json', callback=cb)

        self.pipeline.step_set.create(
            url='http://example.org/step/',
            name='step1',
            inputs=Step.EMPTY_INPUTS,
            outputs=Step.EMPTY_OUTPUTS)
        self.pipeline.endpoint = 'http://example.org/endpoint/'
        self.pipeline.run({})
        self.assertEqual(len(responses.calls), 2)



import dj_database_url
import os.path

DEBUG = True

BASE_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))), '..')

SECRET_KEY = 'foo'

INSTALLED_APPS = (
    'djangobower',
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'wapipelines',
    'rest_framework',
    'rest_framework.authtoken',
)

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': DEBUG,
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'djangobower.finders.BowerFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
]

DATABASES = {
    'default': dj_database_url.config(
        default='postgres://localhost/wa-pipelines'),
}

ROOT_URLCONF = 'wapipelines.tests.test_urls'

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    'DEFAULT_PERMISSION_CLASSES': [
        'rest_framework.permissions.IsAuthenticated',
    ],
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    ),
}

BOWER_COMPONENTS_ROOT = BASE_DIR
BOWER_PATH = os.path.join(BASE_DIR, 'node_modules', '.bin', 'bower')
BOWER_INSTALLED_APPS = (
    'bootstrap#3.3.7',
    'viz-js#0.0.4',
)

SECOND = 1
MINUTE = 60 * SECOND
HOUR = 60 * MINUTE
DAY = 24 * HOUR

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')

MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')




from django.conf.urls import url, include
from django.conf.urls.static import static
from django.contrib import admin
from django.conf import settings

from rest_framework import routers

from wapipelines.api import views

router = routers.SimpleRouter()
router.register(r'pipeline', views.PipelineViewSet, 'pipeline')


urlpatterns = [
    url(r'^', include('wapipelines.urls', namespace='pipeline')),
    url(r'^api/', include(router.urls, namespace='api')),
    url(r'^admin/', admin.site.urls),
] + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-09-25 08:56
from __future__ import unicode_literals

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid
import wapipelines.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Pipeline',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('uuid', models.UUIDField(default=uuid.uuid4)),
                ('secret', models.CharField(blank=True, default=wapipelines.models.generate_secret, max_length=255)),
                ('planning', django.contrib.postgres.fields.ArrayField(base_field=models.UUIDField(blank=True, null=True), default=[], size=None)),
                ('dot', models.TextField(blank=True, null=True)),
                ('endpoint', models.URLField(blank=True, null=True)),
                ('active', models.BooleanField(default=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='PipelineRun',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4)),
                ('dot', models.TextField(blank=True, null=True)),
                ('expected_outputs', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), size=None)),
                ('payload', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('pipeline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wapipelines.Pipeline')),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4)),
                ('inputs', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('outputs', django.contrib.postgres.fields.jsonb.JSONField(default=dict)),
                ('duration', models.DurationField()),
                ('response_code', models.PositiveIntegerField(blank=True, null=True)),
                ('error', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('pipeline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wapipelines.Pipeline')),
                ('run', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wapipelines.PipelineRun')),
            ],
            options={
                'ordering': ('created_at',),
            },
        ),
        migrations.CreateModel(
            name='Step',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(blank=True, max_length=255, null=True)),
                ('uuid', models.UUIDField(default=uuid.uuid4)),
                ('inputs', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=125), size=None)),
                ('outputs', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=125), size=None)),
                ('url', models.URLField(blank=True, null=True)),
                ('headers', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name=b'HTTP headers to submit as part of the step')),
                ('timeout', models.PositiveIntegerField(choices=[(50, b'50 milliseconds'), (100, b'100 milliseconds'), (200, b'200 milliseconds'), (3000, b'3 seconds'), (10000, b'10 seconds')], default=100, verbose_name=b'Timeout in milliseconds')),
                ('error', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('pipeline', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wapipelines.Pipeline')),
            ],
        ),
        migrations.AddField(
            model_name='result',
            name='step',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='wapipelines.Step'),
        ),
    ]



from celery import shared_task
from celery.utils.log import get_task_logger

from requests import HTTPError


logger = get_task_logger(__name__)


@shared_task(autoretry_for=(HTTPError,),
             retry_kwargs={'max_retries': 5})
def run_step_task(run_pk, step_pk, payload, inputs):
    from wapipelines.models import Step, PipelineRun
    step = Step.objects.get(pk=step_pk)
    pipeline = step.pipeline
    run = PipelineRun.objects.get(pk=run_pk)
    logger.info('Running step %s for %s with inputs %s' % (
        step.name, run.uuid, inputs.keys()))

    # NOTE: Because of transaction management, this may return
    #       True for two tasks running at the same time
    #       resulting in tasks doing the same things twice.
    #
    #       For now that is OK.
    if run.still_needs(step):
        step.run_step(run, payload, inputs)
    outputs = step.pipeline.get_outputs(run)
    [run_step_task.delay(run.pk, next_step.pk, payload, outputs)
     for next_step in step.next_steps(run=run)
     if next_step.is_ready(outputs)]

    if step.is_exit_point() and pipeline.endpoint:
        pipeline_submit_outputs.delay(run.pk, step.pk)


@shared_task(autoretry_for=(HTTPError,),
             retry_kwargs={'max_retries': 5})
def pipeline_submit_outputs(run_pk, step_pk):
    from wapipelines.models import Step, PipelineRun
    step = Step.objects.get(pk=step_pk)
    pipeline = step.pipeline
    run = PipelineRun.objects.get(pk=run_pk)
    pipeline.submit_outputs(run, step)



import logging
import six
import json

from django import template
from django.core.serializers.json import DjangoJSONEncoder


logger = logging.getLogger(__file__)

register = template.Library()


@register.filter
def json_dump(object):
    return json.dumps(object, indent=2, cls=DjangoJSONEncoder)


# NOTE: from https://stackoverflow.com/a/1242107
@register.filter
def oxford_join(obj_list):
    """Takes a list of objects and returns their unicode representations,
    seperated by commas and with 'and' between the penultimate and final items
    For example, for a list of fruit objects:
    [<Fruit: apples>,<Fruit: oranges>,<Fruit: pears>] ->
     'apples, oranges, and pears'
    """
    if not obj_list:
        return ""
    l = len(obj_list)
    if l == 1:
        return u"%s" % obj_list[0]
    else:
        return (", ".join(six.u(obj) for obj in obj_list[:l - 1]) +
                ", and " + six.u(obj_list[l - 1]))



import responses
import json
from uuid import UUID

from django.shortcuts import reverse
from django.contrib.auth.models import User

from rest_framework.test import APITestCase, APIClient
from rest_framework.authtoken.models import Token

from wapipelines.models import Pipeline, PipelineRun, Step

from mock import patch

ARTEFACT1 = 'artefact-1'


class PipelineAPITestCase(APITestCase):

    DEFAULT_USERNAME = 'username'

    def setUp(self):
        self.user = self.get_user(
            username='view-username', password='password')
        self.user.is_superuser = True
        self.user.save()

        self.pipeline = Pipeline.objects.create(
            user=self.user,
            name='test pipeline')

        self.client = APIClient()
        self.client.credentials(
            HTTP_AUTHORIZATION='Token ' + self.user.auth_token.key)

    def get_user(self, username=None, email_address='user@example.org',
                 password=None):
        user, _ = User.objects.get_or_create(
            username=(username or self.DEFAULT_USERNAME), email=email_address)
        if password:
            user.set_password(password)
        user.save()
        Token.objects.update_or_create(user=user, defaults={})
        return user

    @responses.activate
    def test_run_pipeline_bad_request(self):
        response = self.client.post(reverse('api:pipeline-run', kwargs={
            'uuid': self.pipeline.uuid,
        }))
        self.assertEqual(response.json(), {
            "payload": ["This field is required."],
            "inputs": ["This field is required."],
        })

    @responses.activate
    @patch('wapipelines.tasks.run_step_task')
    def test_run_pipeline_good_request(self, mocked_task):
        step = self.pipeline.step_set.create(
            name='step1', url='http://example.org/',
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT1])

        response = self.client.post(reverse('api:pipeline-run', kwargs={
            'uuid': self.pipeline.uuid,
        }), data=json.dumps({
            'payload': {'foo': 'bar'},
            'inputs': {},
        }), content_type='application/json')
        data = response.json()
        self.assertEqual(list(data.keys()), ['run'])
        self.assertTrue(UUID(data['run']))
        run = PipelineRun.objects.get(uuid=data['run'])
        mocked_task.delay.assert_called_with(
            run.pk, step.pk, {'foo': 'bar'}, {})

    @responses.activate
    @patch('wapipelines.tasks.run_step_task')
    def test_run_pipeline_good_direct(self, mocked_task):
        step = self.pipeline.step_set.create(
            name='step1', url='http://example.org/',
            inputs=Step.EMPTY_INPUTS,
            outputs=[ARTEFACT1])

        response = self.client.post(
            reverse('api:pipeline-direct', kwargs={
                'uuid': self.pipeline.uuid,
            }),
            data=json.dumps({'foo': 'bar'}),
            content_type='application/json')
        data = response.json()
        self.assertEqual(list(data.keys()), ['run'])
        self.assertTrue(UUID(data['run']))
        run = PipelineRun.objects.get(uuid=data['run'])
        mocked_task.delay.assert_called_with(
            run.pk, step.pk, {'foo': 'bar'}, {})

    @responses.activate
    def test_run_pipeline_good_request_bad_inputs(self):
        self.pipeline.step_set.create(
            name='step1', url='http://example.org/',
            inputs=[ARTEFACT1],
            outputs=[])

        response = self.client.post(reverse('api:pipeline-run', kwargs={
            'uuid': self.pipeline.uuid,
        }), data=json.dumps({
            'payload': {'foo': 'bar'},
            'inputs': {},
        }), content_type='application/json')
        self.assertEqual(response.json(), {
            'inputs': ['No suitable inputs given'],
        })

    @responses.activate
    @patch('wapipelines.tasks.run_step_task')
    def test_run_pipeline_bad_direct(self, mocked_task):
        self.pipeline.step_set.create(
            name='step1', url='http://example.org/',
            inputs=[ARTEFACT1],
            outputs=Step.EMPTY_OUTPUTS)

        response = self.client.post(
            reverse('api:pipeline-direct', kwargs={
                'uuid': self.pipeline.uuid,
            }),
            data=json.dumps({'foo': 'bar'}),
            content_type='application/json')
        self.assertEqual(response.json(), {
            'inputs': ['This pipeline requires inputs.'],
        })



import logging
import json
import os
import six
import hmac
import hashlib
import binascii
import pkg_resources

from datetime import datetime
from collections import defaultdict

from django.db import models
from django.contrib.postgres import fields
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

from uuid import uuid4
import requests
import networkx as nx
from networkx.algorithms.dag import topological_sort
from networkx.drawing import nx_pydot

logger = logging.getLogger(__name__)


class PipelineException(Exception):
    pass


class ImpossiblePipelineException(PipelineException):
    pass


class PipelineStepException(PipelineException):
    pass


def run_step_sync(run, step, payload, inputs):
    logger.info('Running %s' % (step.name,))
    pipeline = step.pipeline
    result = step.run_step(run, payload, inputs)
    outputs = step.pipeline.get_outputs(run)
    next_steps = [run_step_sync(run, next_step, payload, outputs)
                  for next_step in step.next_steps(run=run)
                  if next_step.is_ready(outputs)]

    if step.is_exit_point() and pipeline.endpoint:
        pipeline.submit_outputs(run, step)

    return (step, result, next_steps)


def run_step_async(run, step, payload, inputs):
    logger.info('Scheduling %s' % (step.name,))
    from wapipelines.tasks import run_step_task
    return run_step_task.delay(run.pk, step.pk, payload, inputs)


SYNC_RUNNER = run_step_sync
ASYNC_RUNNER = run_step_async


def generate_secret():
    return binascii.hexlify(os.urandom(16))


class Pipeline(models.Model):
    user = models.ForeignKey('auth.User')
    name = models.CharField(max_length=255)
    uuid = models.UUIDField(default=uuid4)
    secret = models.CharField(
        default=generate_secret,
        null=False, blank=True, max_length=255)
    planning = fields.ArrayField(
        models.UUIDField(blank=True, null=True), default=[])
    dot = models.TextField(blank=True, null=True)
    endpoint = models.URLField(blank=True, null=True)
    active = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    EMPTY_INPUTS = {}

    @classmethod
    def graph(cls, pipeline):
        dag = nx.DiGraph()
        steps = pipeline.step_set.all().order_by('created_at')
        for step in steps:
            dag.add_node(step)

        for current_step in steps:
            for next_step in current_step.next_steps():
                dag.add_edge(current_step, next_step)

            for previous_step in current_step.previous_steps():
                dag.add_edge(previous_step, current_step)

        return dag

    @classmethod
    def plan(cls, pipeline):
        try:
            graph = Pipeline.graph(pipeline)
            topological_sort(graph)
            return graph
        except (nx.NetworkXUnfeasible,):
            raise ImpossiblePipelineException()

    @classmethod
    def find_cycles(self, pipeline):
        return nx.simple_cycles(Pipeline.graph(pipeline))

    def hexdigest(self, data, secret=None):
        return hmac.new(
            key=six.b(str(secret or self.secret)),
            msg=six.b(str(data)),
            digestmod=hashlib.sha512).hexdigest()

    def get_default_headers(self, data):
        return {
            'Content-Type': 'application/json',
            'User-Agent': 'WA-Pipeline/%s' % (
                pkg_resources.get_distribution('wapipelines').version),
            'X-WA-Pipeline': self.uuid.hex,
            'X-WA-Sign': self.hexdigest(data),
        }

    def signed_http_post(self, url, data=None, timeout=None,
                         headers={}):
        default_headers = self.get_default_headers(data)
        default_headers.update(headers)
        return requests.post(
            url,
            data=data,
            timeout=timeout,
            headers=default_headers)

    def submit_outputs(self, run, step):
        r = self.signed_http_post(
            self.endpoint,
            data=json.dumps({
                'run': run.serialize(),
                'pipeline': self.serialize(run),
            }, indent=2, cls=DjangoJSONEncoder),
            timeout=(1 * settings.MINUTE),
            headers={
                'X-WA-Step': step.uuid.hex,
                'X-WA-Step-Name': step.name,
            })
        r.raise_for_status()
        return r

    def get_cycles(self):
        return list(Pipeline.find_cycles(self))

    def to_pydot(self):
        return nx_pydot.to_pydot(Pipeline.graph(self))

    def set_plan(self, graph):
        self.dot = nx_pydot.to_pydot(graph).to_string()

    def can_receive(self, other_pipeline):
        connections = []
        for entry_point in self.entry_points():
            for exit_point in other_pipeline.exit_points():
                if entry_point.inputs == exit_point.outputs:
                    connections.append((exit_point, entry_point))
        return connections

    def entry_points(self):
        graph = Pipeline.graph(self)
        return [step for step in graph.nodes()
                if not graph.in_edges(step)]

    def exit_points(self):
        graph = Pipeline.graph(self)
        return [step for step in graph.nodes()
                if not graph.out_edges(step)]

    def can_run_given(self, inputs):
        if not self.active:
            return False
        return inputs in self.allowable_inputs()

    def run(self, payload, inputs=None, runner=None):
        runner = runner or SYNC_RUNNER

        inputs = inputs or self.EMPTY_INPUTS
        run = self.pipelinerun_set.create(
            payload=payload, dot=self.dot,
            expected_outputs=list(self.possible_outputs()))
        steps = self.step_set.filter(
            models.Q(inputs__overlap=list(inputs.keys())) |
            models.Q(inputs__len=0))

        return (run, [runner(run, step, payload, inputs)
                      for step in steps])

    def requires_inputs(self):
        return any([inputs for inputs in self.allowable_inputs()])

    def allowable_inputs(self):
        inputs = []
        # set struggles to make sets of lists of lists
        for step in self.entry_points():
            if step.inputs not in inputs:
                inputs.append(step.inputs)
        return inputs

    def possible_outputs(self):
        return set([
            output
            for step in self.step_set.all()
            for output in step.outputs])

    def get_outputs(self, run):
        result_set = defaultdict(list)
        for result in self.result_set.filter(run=run):
            for key, value in result.outputs.items():
                result_set[key].append(value)
        return dict(result_set)

    def serialize(self, run):
        return {
            'uuid': self.uuid,
            'name': self.name,
            'outputs': self.get_outputs(run)
        }


class Step(models.Model):

    MIN_TIMEOUT = 50
    DEFAULT_TIMEOUT = 100
    ACCEPTABLE_TIMEOUT = 200
    MAX_TIMEOUT = 3000
    GOOGLE_CLOUD_TIMEOUT = 10000

    TIMEOUT_CHOICES = (
        (MIN_TIMEOUT, '50 milliseconds'),
        (DEFAULT_TIMEOUT, '100 milliseconds'),
        (ACCEPTABLE_TIMEOUT, '200 milliseconds'),
        (MAX_TIMEOUT, '3 seconds'),
        (GOOGLE_CLOUD_TIMEOUT, '10 seconds'),
    )

    EMPTY_INPUTS = []
    EMPTY_OUTPUTS = []

    pipeline = models.ForeignKey(Pipeline)
    name = models.CharField(max_length=255, blank=True, null=True)
    uuid = models.UUIDField(default=uuid4)
    inputs = fields.ArrayField(
        models.CharField(max_length=125, blank=True))
    outputs = fields.ArrayField(
        models.CharField(max_length=125, blank=True))
    url = models.URLField(null=True, blank=True)
    headers = fields.JSONField(
        'HTTP headers to submit as part of the step',
        default=dict)
    timeout = models.PositiveIntegerField(
        'Timeout in milliseconds',
        default=DEFAULT_TIMEOUT,
        choices=TIMEOUT_CHOICES)
    error = models.TextField(blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def get_missing_inputs(self, inputs):
        return set(self.inputs) - set(inputs.keys())

    def is_ready(self, inputs):
        return not any(self.get_missing_inputs(inputs))

    def is_noop(self):
        """
        Some steps have no other purpose but to wait for results
        for preceding steps, currently the only way of specifying
        that behaviour is by leaving the URL blank
        """
        return not self.url

    def get_default_headers(self, run):
        return {
            'X-WA-Run': run.uuid.hex,
            'X-WA-Step': self.uuid.hex,
            'X-WA-Step-Name': self.name,
        }

    def run_step(self, run, payload, inputs):
        if self.is_noop():
            return

        start = datetime.now()
        seconds_timeout = (self.timeout / 1000.0)
        response = None
        headers = self.get_default_headers(run)
        headers.update(self.headers)
        try:
            response = self.pipeline.signed_http_post(
                url=self.url,
                data=json.dumps({
                    "payload": payload,
                    "inputs": inputs,
                }, indent=2, cls=DjangoJSONEncoder),
                timeout=seconds_timeout,
                headers=headers
            )
            outputs = response.json()
            response.raise_for_status()
            return self.result_set.create(
                pipeline=self.pipeline,
                run=run,
                inputs=inputs, outputs=outputs,
                duration=datetime.now() - start,
                response_code=response.status_code)
        except (requests.RequestException, ValueError) as e:
            response_code = (response.status_code
                             if response is not None
                             else None)
            return self.result_set.create(
                pipeline=self.pipeline,
                run=run,
                inputs=inputs, outputs={},
                duration=datetime.now() - start,
                response_code=response_code,
                error=str(e))

    def next_steps(self, run=None):
        steps = Step.objects.filter(
            pipeline=self.pipeline, inputs__overlap=self.outputs
        ).order_by('created_at')
        if run is not None:
            steps = steps.exclude(result__run=run)
        return steps

    def previous_steps(self):
        return Step.objects.filter(
            pipeline=self.pipeline, outputs__overlap=self.inputs
        ).order_by('created_at')

    def is_exit_point(self):
        return self in self.pipeline.exit_points()

    def __str__(self):
        return self.name or str(self.uuid)


class PipelineRun(models.Model):
    pipeline = models.ForeignKey(Pipeline)
    uuid = models.UUIDField(default=uuid4)
    dot = models.TextField(null=True, blank=True)
    expected_outputs = fields.ArrayField(
        models.CharField(max_length=255)
    )
    payload = fields.JSONField(null=True, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ('-created_at',)

    def still_needs(self, step):
        return not self.result_set.filter(step=step).exists()

    def serialize(self):
        return {
            'uuid': self.uuid,
            'is_completed': self.is_completed(),
            'expected_outputs': self.expected_outputs,
            'created_at': self.created_at,
            'payload': self.payload,
            'errors': [
                result.serialize() for result in self.get_errors()],
            'results': [
                result.serialize() for result in self.result_set.all()]
        }

    def get_errors(self):
        return self.result_set.filter(error__isnull=False)

    def is_completed(self):
        if self.get_errors():
            return False

        return set(self.get_outputs().keys()).issuperset(
            set(self.expected_outputs))

    def get_outputs(self):
        return self.pipeline.get_outputs(self)

    def duration(self):
        aggregate = self.result_set.aggregate(
            duration=(models.Max('created_at') - models.Min('created_at')))
        return aggregate.get('duration')

    def __str__(self):
        return self.uuid.hex


class Result(models.Model):
    pipeline = models.ForeignKey(Pipeline)
    uuid = models.UUIDField(default=uuid4)
    run = models.ForeignKey(PipelineRun)
    step = models.ForeignKey(Step)
    inputs = fields.JSONField(default=dict)
    outputs = fields.JSONField(default=dict)
    duration = models.DurationField()
    response_code = models.PositiveIntegerField(null=True, blank=True)
    error = models.TextField(blank=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def serialize(self):
        return {
            'uuid': self.uuid,
            'created_at': self.created_at,
            'inputs': self.inputs,
            'outputs': self.outputs,
            'step_uuid': self.step.uuid,
            'step_name': self.step.name,
            'duration': self.duration.total_seconds(),
            'error': self.error,
        }

    def get_step_headers(self):
        default_headers = self.pipeline.get_default_headers(
            json.dumps(self.get_step_data(), cls=DjangoJSONEncoder))
        default_headers.update(self.step.get_default_headers(self.run))
        return default_headers

    def get_step_data(self):
        return {
            "payload": self.run.payload,
            "inputs": self.inputs,
        }

    class Meta:
        ordering = ('created_at',)



import json
import email

from django import forms
from django.forms import widgets
from django.contrib.postgres import forms as pg_forms

from wapipelines.models import Pipeline, Step


class HTTPHeaderField(forms.Field):

    def clean(self, data):
        if data and ':' not in data:
            raise forms.ValidationError(
                'Headers must be in Key: Value format.')
        message = email.message_from_string(data)
        return dict([
            (key, message.get(key))
            for key in message.keys()])


class HTTPHeaderWidget(forms.Textarea):

    def format_value(self, raw_value):
        if not raw_value:
            return ''
        if isinstance(raw_value, basestring):
            return raw_value

        return '\n'.join([
            '%s: %s' % (key, value)
            for key, value in raw_value.items()
        ])


class CreatePipelineForm(forms.ModelForm):
    name = forms.CharField(
        widget=widgets.TextInput(attrs={'class': 'form-control'}))
    endpoint = forms.URLField(
        widget=widgets.TextInput(attrs={'class': 'form-control'}))

    class Meta:
        model = Pipeline
        fields = (
            'name',
            'active',
            'endpoint',
        )


class CreateStepForm(forms.ModelForm):
    name = forms.CharField(
        widget=widgets.TextInput(attrs={'class': 'form-control'}))
    url = forms.URLField(
        widget=widgets.TextInput(attrs={'class': 'form-control'}),
        help_text=('Specify a URL to request outputs from, leave blank if '
                   'this step is only going to be used for waiting for '
                   'outputs from preceding steps'),
        required=False)
    timeout = forms.CharField(
        widget=widgets.Select(attrs={'class': 'form-control'},
                              choices=Step.TIMEOUT_CHOICES))
    inputs = pg_forms.SimpleArrayField(
        forms.CharField(),
        widget=widgets.TextInput(attrs={'class': 'form-control'}),
        required=False)
    outputs = pg_forms.SimpleArrayField(
        forms.CharField(),
        widget=widgets.TextInput(attrs={'class': 'form-control'}),
        required=False)
    headers = HTTPHeaderField(
        widget=HTTPHeaderWidget(attrs={'class': 'form-control'}),
        required=False)

    class Meta:
        model = Step
        fields = (
            'name',
            'url',
            'timeout',
            'inputs',
            'outputs',
            'headers',
        )


class PipelineRunForm(forms.Form):
    data = forms.CharField(
        widget=forms.Textarea(attrs={'class': 'form-control'}),
        required=True)

    def clean_data(self):
        raw_data = self.cleaned_data['data']
        if not raw_data:
            raise forms.ValidationError('A JSON value is required')
        try:
            return json.loads(raw_data)
        except (ValueError,):
            raise forms.ValidationError('Invalid JSON')



from rest_framework import serializers
from wapipelines.models import Pipeline


class PipelineSerializer(serializers.ModelSerializer):
    class Meta:
        model = Pipeline
        fields = ('uuid', 'name', 'endpoint')


class RunPipelineSerializer(serializers.Serializer):
    payload = serializers.JSONField()
    inputs = serializers.DictField()



from django.conf.urls import url
from django.contrib.auth.decorators import login_required
from wapipelines import views

urlpatterns = [
    url(r'^$',
        login_required(views.PipelineListView.as_view()),
        name='list'),
    url(r'^create/$',
        login_required(views.PipelineCreateView.as_view()),
        name='create'),
    url(r'^(?P<pk>\d+)/update/$',
        login_required(views.PipelineUpdateView.as_view()),
        name='update'),
    url(r'^(?P<pk>\d+)/run/$',
        login_required(views.PipelineRunView.as_view()),
        name='run'),
    url(r'^(?P<pipeline_pk>\d+)/$',
        login_required(views.StepListView.as_view()),
        name='step_list'),
    url(r'^(?P<pipeline_pk>\d+)/create/$',
        login_required(views.StepCreateView.as_view()),
        name='step_create'),
    url(r'^(?P<pipeline_pk>\d+)/steps/(?P<pk>\d+)/$',
        login_required(views.StepUpdateView.as_view()),
        name='step_update'),
    url(r'^(?P<pipeline_pk>\d+)/steps/(?P<pk>\d+)/delete/$',
        login_required(views.StepDeleteView.as_view()),
        name='step_delete'),
]



from __future__ import absolute_import, unicode_literals
import os
from celery import Celery

os.environ.setdefault(
    'DJANGO_SETTINGS_MODULE', 'wapipelines.tests.test_settings')

app = Celery('wapipelines')
app.config_from_object('django.conf:settings')
app.autodiscover_tasks()



import logging
import json
from django.views.generic.base import View
from django.views.generic import ListView, CreateView, UpdateView, DeleteView
from django.views.generic.edit import FormView

from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, reverse
from django.conf import settings
from django.contrib import messages
from random import randint
from time import sleep

from wapipelines.models import Pipeline, Step, ASYNC_RUNNER
from wapipelines.forms import (
    CreatePipelineForm, CreateStepForm, PipelineRunForm)

from digg_paginator import DiggPaginator


logger = logging.getLogger(__name__)


class PipelineListView(ListView):

    model = Pipeline

    def get_queryset(self, *args, **kwargs):
        qs = super(PipelineListView, self).get_queryset(*args, **kwargs)
        return qs.filter(user=self.request.user)


class PipelineCreateView(CreateView):
    model = Pipeline
    form_class = CreatePipelineForm
    ordering = ('created_at',)

    def form_valid(self, form):
        pipeline = form.save(commit=False)
        pipeline.user = self.request.user
        pipeline.save()
        return redirect('pipeline:list')


class PipelineUpdateView(UpdateView):
    model = Pipeline
    form_class = CreatePipelineForm

    def form_valid(self, form):
        pipeline = form.save(commit=False)
        pipeline.user = self.request.user
        pipeline.save()
        return redirect('pipeline:list')


class StepListView(ListView):
    model = Step
    form_class = CreateStepForm
    paginate_by = 10
    paginator_class = DiggPaginator
    ordering = ('-created_at',)

    def get_context_data(self, *args, **kwargs):
        pipeline = get_object_or_404(
            Pipeline, user=self.request.user,
            pk=self.kwargs['pipeline_pk'])
        context = super(StepListView, self).get_context_data(*args, **kwargs)
        context.update({
            'pipeline': pipeline,
            'runs': pipeline.pipelinerun_set.order_by('-created_at')[:100],
            'form': self.form_class(initial={
                'pipeline': pipeline,
            }),
            'run_form': PipelineRunForm(),
        })
        return context

    def get_queryset(self, *args, **kwargs):
        return Step.objects.filter(
            pipeline__pk=self.kwargs['pipeline_pk'],
            pipeline__user=self.request.user).order_by('created_at')


class StepCreateView(CreateView):
    model = Step
    form_class = CreateStepForm

    def form_valid(self, form):
        step = form.save(commit=False)
        step.pipeline = get_object_or_404(
            Pipeline,
            user=self.request.user,
            pk=self.kwargs['pipeline_pk'])
        step.save()
        return redirect(
            'pipeline:step_list', pipeline_pk=step.pipeline.pk)


class StepUpdateView(UpdateView):
    model = Step
    form_class = CreateStepForm

    def form_valid(self, form):
        step = form.save(commit=False)
        step.pipeline = get_object_or_404(
            Pipeline,
            user=self.request.user,
            pk=self.kwargs['pipeline_pk'])
        step.save()
        return redirect(
            'pipeline:step_list', pipeline_pk=step.pipeline.pk)


class StepDeleteView(DeleteView):
    model = Step

    def get_queryset(self, *args, **kwargs):
        pipeline = get_object_or_404(
            self.request.user.pipeline_set, pk=self.kwargs['pipeline_pk'])
        return pipeline.step_set.all()

    def get_success_url(self):
        pipeline = get_object_or_404(
            self.request.user.pipeline_set, pk=self.kwargs['pipeline_pk'])
        return reverse('pipeline:step_list', kwargs={
            'pipeline_pk': pipeline.pk,
        })


class PipelineRunView(FormView):
    template_name = 'wapipelines/pipeline_run.html'
    form_class = PipelineRunForm

    def form_valid(self, form):
        pipeline = get_object_or_404(
            self.request.user.pipeline_set.all(), pk=self.kwargs['pk'])
        data = form.cleaned_data['data']
        pipeline.run(data, runner=ASYNC_RUNNER)
        messages.info(self.request, 'Pipeline run started')
        return redirect(
            'pipeline:step_list', pipeline_pk=pipeline.pk)


class ArtefactView(View):

    def get(self, request, artefact):
        return JsonResponse({
            "inputs": request.GET.getlist('inputs'),
            "outputs": [artefact],
        }, json_dumps_params={
            'indent': 2,
        })

    def post(self, request, artefact):
        logger.info('Received payload: %s' % (json.load(request),))
        naptime = randint(1, 2) if settings.DEBUG else 0
        logger.info('napping %s for %s seconds' % (artefact, naptime,))
        sleep(naptime)
        return JsonResponse({
            artefact: 'value of %s' % (artefact,),
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from setuptools import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as req_file:
    requirements = req_file.read().split('\n')

with open('requirements-dev.txt') as req_file:
    requirements_dev = req_file.read().split('\n')

with open('VERSION') as fp:
    version = fp.read().strip()

setup(
    name='wapipelines',
    version=version,
    description="WA Pipelines",
    long_description=readme,
    author="Simon de Haan",
    author_email='simon@praekelt.org',
    url='https://github.com/praekeltfoundation/wa-pipelines',
    packages=[
        'wapipelines',
    ],
    package_dir={'wapipelines':
                 'wapipelines'},
    extras_require={
        'dev': requirements_dev,
    },
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='wa-pipelines',
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
    ]
)



from rest_framework import viewsets
from rest_framework import status
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from wapipelines.models import Pipeline, ASYNC_RUNNER
from wapipelines.api.serializers import (
    PipelineSerializer, RunPipelineSerializer)


class PipelineViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = PipelineSerializer
    lookup_field = 'uuid'

    def get_queryset(self):
        return Pipeline.objects.filter(user=self.request.user)

    @detail_route(methods=['post'])
    def run(self, request, uuid=None):
        pipeline = self.get_object()
        serializer = RunPipelineSerializer(data=request.data)
        if serializer.is_valid():
            payload = serializer.data['payload']
            inputs = serializer.data.get('inputs', {})
            # NOTE: list for py3 because it returns a dict_keys() type
            if not pipeline.can_run_given(list(inputs.keys())):
                return Response({
                    'inputs': ['No suitable inputs given']
                }, status=status.HTTP_400_BAD_REQUEST)

            run, first_steps = pipeline.run(
                payload,
                inputs=inputs,
                runner=ASYNC_RUNNER)
            return Response({
                'run': run.uuid,
            }, status=status.HTTP_202_ACCEPTED)
        else:
            return Response(serializer.errors,
                            status=status.HTTP_400_BAD_REQUEST)

    @detail_route(methods=['post'])
    def direct(self, request, uuid=None):
        pipeline = self.get_object()
        if not pipeline.can_run_given([]):
            return Response({
                'inputs': ['This pipeline requires inputs.']
            }, status=status.HTTP_400_BAD_REQUEST)

        run, first_steps = pipeline.run(
            request.data, inputs=[], runner=ASYNC_RUNNER)
        return Response({
            'run': run.uuid,
        }, status=status.HTTP_202_ACCEPTED)
